#include "FiniteElemSpace.h"
#include "DOFIndexed.h"
#include "DOFVector.h"
#include "Traverse.h"
#include "BasisFunction.h"
#include "ElInfo.h"
#include "BoundaryCondition.h"
#include <list>

namespace AMDiS {

  template<typename T>
  double BoundaryManager<T>::boundResidual(ElInfo *elInfo, Estimator<T> *estimator)
  {
    double result = 0;
    typename std::list<LocalBC<T>*>::iterator it;
    for(it = localBCs.begin(); it != localBCs.end(); ++it) {
      result += (*it)->boundResidual(elInfo, estimator);
    }
    return result;
  }

  template<typename T>
  void BoundaryManager<T>::fillGlobalBoundVector(DOFVector<BoundaryType> *globalBound)
  {
    int i;
    TraverseStack  stack;
    const FiniteElemSpace *feSpace = globalBound->getFESpace();

    const BasisFunction *basisFcts = feSpace->getBasisFcts();
    int                  nBasFcts = basisFcts->getNumber();
    Mesh                *mesh = feSpace->getMesh();
    DOFAdmin            *admin = feSpace->getAdmin();

    const BoundaryType *localBound = NULL;
    const DegreeOfFreedom *dofIndices = NULL;

    ElInfo *elInfo = 
      stack.traverseFirst(mesh, -1, 
			  Mesh::CALL_LEAF_EL | 
			  Mesh::FILL_BOUND | 
			  Mesh::FILL_COORDS);

    // for all elements ...
    while(elInfo) {
      // get boundaries of all DOFs
      localBound = basisFcts->getBound(elInfo, NULL);
    
      // get dof indices
      dofIndices = basisFcts->getLocalIndices(elInfo->getElement(),
					      admin, NULL);

      // for all DOFs
      for(i=0; i < nBasFcts; i++) {
	(*globalBound)[dofIndices[i]] = localBound[i];
      }

      elInfo = stack.traverseNext(elInfo);
    }
  }

  template<typename T>
  void BoundaryManager<T>::fillLocalBCs(ElInfo *elInfo, 
					DOFVector<T> *vec)
  {
    int i;

    // ===== fill local conditions ==============================================
    const FiniteElemSpace *feSpace = vec->getFESpace();
    const BoundaryType    *localBound = NULL;
    const DegreeOfFreedom *dofIndices = NULL;
    Mesh                  *mesh = feSpace->getMesh();
    const BasisFunction   *basisFcts = feSpace->getBasisFcts();
    int                    nBasFcts = basisFcts->getNumber();
    LocalBC<T>            *localBC;
    DOFAdmin              *admin = feSpace->getAdmin();

    typename std::list<LocalBC<T>*>::iterator it;

    if(localBCs.size() > 0) {

      // get boundaries of all DOFs
      localBound = basisFcts->getBound(elInfo, NULL);

      // get dof indices
      dofIndices = basisFcts->getLocalIndices(elInfo->getElement(),
					      admin, NULL);

      // apply boundary conditions
      for(it = localBCs.begin(); it != localBCs.end(); ++it) {
	(*it)->fillLocalBC(vec, elInfo, dofIndices, localBound, nBasFcts);
      }
    }
  }

  template<typename T>
  void BoundaryManager<T>::fillLocalBCs(ElInfo *elInfo, 
					DOFMatrix *mat)
  {
    int i;

    // ===== fill local conditions ==============================================
    const FiniteElemSpace *feSpace = mat->getRowFESpace();
    const BoundaryType    *localBound = NULL;
    const DegreeOfFreedom *dofIndices = NULL;
    Mesh                  *mesh = feSpace->getMesh();
    const BasisFunction   *basisFcts = feSpace->getBasisFcts();
    int                    nBasFcts = basisFcts->getNumber();
    LocalBC<T>            *localBC;
    DOFAdmin              *admin = feSpace->getAdmin();

    typename std::list<LocalBC<T>*>::iterator it;

    if(localBCs.size() > 0) {

      // get boundaries of all DOFs
      localBound = basisFcts->getBound(elInfo, NULL);

      // get dof indices
      dofIndices = basisFcts->getLocalIndices(elInfo->getElement(),
					      admin, NULL);

      // apply boundary conditions
      for(it = localBCs.begin(); it != localBCs.end(); ++it) {
	(*it)->fillLocalBC(mat, elInfo, dofIndices, localBound, nBasFcts);
      }
    }
  }

}
