#include "GMResSolver2.h"
#include "Preconditioner.h"

namespace AMDiS {

  template<typename VectorType>
  GMResSolver2<VectorType>::GMResSolver2(std::string name)
    : OEMSolver<VectorType>(name),
      TOL_(1.e-25),
      restart_(0),
      w_(NULL)
  {
    FUNCNAME("GMResSolver2::GMResSolver2()");
    GET_PARAMETER(0, name + "->restart", "%d", &restart_);
    TEST_EXIT(restart_)("restart not set\n");
  }

  template<typename VectorType>
  GMResSolver2<VectorType>::~GMResSolver2()
  {}

  template<typename VectorType>
  void GMResSolver2<VectorType>::init()
  {
    // Create a new vector w.
    w_ = this->vectorCreator->create();
    z_ = this->vectorCreator->create();
    r_ = this->vectorCreator->create();

    // Get memory for the vector pointers and create the pointers afterwards.
    v = GET_MEMORY(VectorType*, restart_ + 1);
    for (int i = 0; i <= restart_; i++) {
      v[i] = this->vectorCreator->create();
    }

    // Resize all fields to the corresponding size.
    gamma.resize(restart_ + 1);
    c.resize(restart_);
    s.resize(restart_);

    y_.resize(restart_);

    h.resize(restart_ + 1);
    for (int i = 0; i <= restart_; i++) {
      h[i].resize(restart_);
    }
  }

  template<typename VectorType>
  void GMResSolver2<VectorType>::exit()
  {
    // Empty used memory.
    if (w_) {      
      this->vectorCreator->free(w_);
      this->vectorCreator->free(z_);
      this->vectorCreator->free(r_);

      for (int i = 0; i <= restart_; i++) {
	this->vectorCreator->free(v[i]);
      }
    }
  }


  template<typename VectorType>
  int GMResSolver2<VectorType>::gmres(MatVecMultiplier<VectorType> *matVec,
				      VectorType *x, 
				      VectorType *b)
  {
    FUNCNAME("GMResSolver2::gmres()");

    // r_0 = b - Ax, where r_0 is already stored as the first vector in the
    // matrix V.
    matVec->matVec(NoTranspose, *x, *v[0]);
    xpay(-1.0, *b, *v[0]);

    // Left preconditioning, if required.
    if (this->leftPrecon) {
      this->leftPrecon->precon(v[0]);
    }

    // Compute the norm of r_0 = v_0.
    gamma[0] = norm(v[0]);

    // Normalize v_0.
    *v[0] *= (1 / gamma[0]);

    // If the norm of gamma_0 is less than the tolerance bounday, x is already
    // a good solution for Ax = b.
    if (gamma[0] < this->tolerance) {
      this->residual = gamma[0];
      return(0);
    }    
    
    // Main loop of the GMRES algorithm.
    for (int j = 0; j < restart_; j++) {
      // w_j = A * v_j;
      matVec->matVec(NoTranspose, *v[j], *w_);
      
      // Preconditioning of w_j
      if (this->leftPrecon) {
	this->leftPrecon->precon(w_);
      }

      // w_j = A * v_j - \sum(h_ij * v_i)
      for (int i = 0; i <= j; i++) {
	// h_ij = (v_i, A * v_j)_2
	h[i][j] = *w_ * (*v[i]);

	// Update w_j
	axpy(-h[i][j], *v[i], *w_);
      }
      
      // h_{j + 1}{j} = \| w_j \|_2
      h[j + 1][j] = norm(w_);
     
      // Compute h_ij and h_{i + 1}{j} using parameters of the Givens rotation.
      for (int i = 0; i < j; i++) {
	double t1 = c[i] * h[i][j] + s[i] * h[i + 1][j];
	double t2 = -s[i] * h[i][j] + c[i] * h[i + 1][j];
	h[i][j] = t1;
	h[i + 1][j] = t2;
      }
      
      // Compute new beta
      double beta = sqrt(h[j][j] * h[j][j] + h[j + 1][j] * h[j + 1][j]);
      // Update s_j and c_j
      s[j] = h[j + 1][j] / beta;
      c[j] = h[j][j] / beta;
      // h_jj = beta
      h[j][j] = beta;

      // Update gammas
      gamma[j + 1] = -s[j] * gamma[j];
      gamma[j] *= c[j];
      
      // Set v_{j + 1} to w_j and normalize it.
      *v[j + 1] = *w_;
      *v[j + 1] *= 1 / h[j + 1][j];      
    }
    
    // Compute the solution.
    for (int i = restart_ - 1; i >= 0; i--) {
      for (int k = i + 1; k < restart_; k++) {
 	gamma[i] -= h[i][k] * gamma[k];
      }
      gamma[i] /= h[i][i];
      axpy(gamma[i], *v[i], *x);
    }

    // Update the new residual.
    this->residual = abs(gamma[restart_]);
       
    return restart_;
  }


  template<typename VectorType>
  int GMResSolver2<VectorType>::solveSystem(MatVecMultiplier<VectorType> *matVec,
 					    VectorType *x, 
 					    VectorType *b)
  {
    FUNCNAME("GMResSolver2::solveSystem()");    
    
    double old_res = -1.0;
    
    // If norm of b is smaller than the tolarance, we can assume b to be zero.
    // Hence, x = 0 is the solution of the linear system.
    if (norm(b) < TOL_) {
      INFO(this->info, 2)("b == 0, x = 0 is the solution of the linear system\n");
      setValue(*x, 0.0);
      this->residual = 0.0;
      return(0);
    }
    
    START_INFO();
    for (int iter = 0; iter <= this->max_iter; iter++) {
      // Solve Ax=b using GMRES(m).
      int k = gmres(matVec, x, b);
      // Check the new residual.
      if (SOLVE_INFO(iter, this->residual, &old_res))
	return(iter);
      TEST_EXIT(k != 0)("this must not happen\n");
    }
  
    ERROR_EXIT("Should not be reached\n");
    return(0);
  }

}

