#ifndef AMDIS_SUB_ASSEMBLER_H
#define AMDIS_SUB_ASSEMBLER_H

#include <map>
#include <vector>
#include "FixVec.h"
#include "Flag.h"
#include "MemoryManager.h"
#include "OpenMP.h"

namespace AMDiS {

  class Assembler;
  class Quadrature;
  class FastQuadrature;
  class Operator;
  class OperatorTerm;
  class BasisFunction;
  class FiniteElemSpace;
  class ElInfo;
  class ElementMatrix;
  class ElementVector;

  template<typename T> class DOFVectorBase;

  /**
   * \ingroup Assembler
   * 
   * \brief
   * Base class for SecondOrderAssembler, FirstOrderAssembler, 
   * ZeroOrderAssembler. The task of a SubAssembler is to assemble a list of 
   * terms of a special order and add their contributions to a DOFMatrix or a 
   * DOFVector. An Assembler can consist of up to four SubAssemblers: one 
   * SecondOrderAssembler for second order terms, one ZeroOrderAssembler for 
   * terms of order zero, and two FirstOrderAssemblers. One for terms with
   * derivatives of the basis functions connected to to row DOFs and one for 
   * those connected to the column DOFs.
   */
  class SubAssembler
  {
  public:
    MEMORY_MANAGED(SubAssembler);

    /** \brief
     * Creates a SubAssembler belonging to assembler for the terms of given 
     * order of Operator op. If order is equal to one, type spezifies what kind 
     * of FirstOrderType are to assemble. During construction of a SubAssembler
     * the needs and properties of the terms are considered.
     */
    SubAssembler(Operator *op,
		 Assembler *assembler,
		 Quadrature *quadrat,
		 int order, 
		 bool optimized,
		 FirstOrderType type = GRD_PHI);

    /** \brief
     * Destructor
     */
    virtual ~SubAssembler() {};

    /** \brief
     * Calculates the element matrix for elInfo and adds it to mat. Memory
     * for mat must be provided by the caller.
     */
    virtual void calculateElementMatrix(const ElInfo *elInfo,
					ElementMatrix *mat) = 0;

    virtual void calculateElementMatrix(const ElInfo *rowElInfo,
					const ElInfo *colElInfo,
					ElementMatrix *mat) = 0;

    /** \brief
     * Calculates the element vector for elInfo and adds it to vec. Memory
     * for vec must be provided by the caller.
     */
    virtual void calculateElementVector(const ElInfo *elInfo,
					ElementVector *vec) = 0;

    /** \brief
     * Returns \ref terms
     */
    inline std::vector<OperatorTerm*> *getTerms() { 
      return &terms[omp_get_thread_num()]; 
    };

    /** \brief
     * Returns \ref quadrature.
     */ 
    inline Quadrature *getQuadrature() {
      return quadrature;
    };

    /** \brief
     * Sets \ref quadrature to q.
     */
    inline void setQuadrature(Quadrature* q) {
      quadrature = q;
    };
  
    /** \brief
     * Returns a vector with the world coordinates of the quadrature points
     * of \ref quadrature on the element of elInfo. 
     * Used by \ref OperatorTerm::initElement().
     */
    WorldVector<double>* getCoordsAtQPs(const ElInfo* elInfo, 
					Quadrature *quad = NULL);

    /** \brief
     * DOFVector dv evaluated at quadrature points.
     * Used by \ref OperatorTerm::initElement().
     */
    double* getVectorAtQPs(DOFVectorBase<double>* dv, const ElInfo* elInfo,
			   Quadrature *quad = NULL);

    /** \brief
     * Gradients of DOFVector dv evaluated at quadrature points.
     * Used by \ref OperatorTerm::initElement().
     */
    WorldVector<double>* getGradientsAtQPs(DOFVectorBase<double>* dv, 
					   const ElInfo* elInfo,
					   Quadrature *quad = NULL);

    /** \brief
     * Called once for each ElInfo when \ref calculateElementMatrix() or
     * \ref calculateElementVector() is called for the first time for this
     * Element.
     */
    virtual void initElement(const ElInfo *elInfo,
			     Quadrature *quad = NULL);

    /** \brief
     * Returns \ref psiFast.
     */
    const FastQuadrature *getPsiFast() const { 
      return psiFast; 
    };

    /** \brief
     * Returns \ref phiFast.
     */
    const FastQuadrature *getPhiFast() const { 
      return phiFast; 
    };

  protected:
    /** \brief
     * Updates \ref psiFast and \ref phiFast.
     */
    FastQuadrature *updateFastQuadrature(FastQuadrature *quadFast,
					 const BasisFunction *psi,
					 Flag updateFlag);
  
  protected:
    /** \brief
     * Problem dimension
     */
    int dim;

    /** \brief
     * Number of rows of the element matrix and length of the element
     * vector. Is equal to the number of row basis functions
     */
    int nRow;

    /** \brief
     * Number of columns of the element matrix. Is equal to the number
     * of column basis functions
     */
    int nCol;

    /** \brief
     * Used for \ref getVectorAtQPs().
     */
    class ValuesAtQPs {
    public:
      Vector<double> values;
      bool valid;
    };

    /** \brief
     * Used for \ref getGradientsAtQPs().
     */
    class GradientsAtQPs {
    public:
      Vector<WorldVector<double> > values;
      bool valid;
    };

    /** \brief
     * Used for \ref getVectorAtQPs().
     */
    std::map<const DOFVectorBase<double>*, ValuesAtQPs*> valuesAtQPs;

    /** \brief
     * Used for \ref getGradientsAtQPs().
     */
    std::map<const DOFVectorBase<double>*, GradientsAtQPs*> gradientsAtQPs;

    /** \brief
     * Set and updated by \ref initElement() for each ElInfo. 
     * coordsAtQPs[i] points to the coordinates of the i-th quadrature point.
     */
    WorldVector<double> *coordsAtQPs;

    /** \brief
     * Used for \ref getCoordsAtQPs().
     */
    bool coordsValid;

    /** \brief
     * Used for \ref getCoordsAtQP(). Stores the number of allocated WorldVectors.
     */
    int coordsNumAllocated;

    /** \brief
     * Needed Quadrature. Constructed in the constructor of SubAssembler
     */
    Quadrature *quadrature;

    /** \brief
     * FastQuadrature for row basis functions
     */
    FastQuadrature *psiFast;

    /** \brief
     * FastQuadrature for column basis functions
     */
    FastQuadrature *phiFast;

    /** \brief
     * Corresponding Assembler
     */
    Assembler* owner;

    /** \brief
     * Flag that specifies whether the element matrix is symmetric.
     */
    bool symmetric;

    /** \brief
     * List of all terms with a contribution to this SubAssembler
     */
    std::vector< std::vector<OperatorTerm*> > terms;

    /** \brief
     *
     */
    bool opt;

    /** \brief
     *
     */
    bool firstCall;

    friend class Assembler;
  };

}

#endif // AMDSIS_SUB_ASSEMBLER_H
